;; -*- Mode: Lisp; Syntax: ANSI-Common-Lisp -*-

(in-package :iolib-posix)

(define-condition system-error (error)
  ((code :reader system-error-code
         :initarg :code
         :type (or null integer))
   (identifier :reader system-error-identifier
               :initarg :identifier
               :type symbol)
   (message :reader message
            :initarg :message)))

(define-condition unix-error (system-error) ())

(defparameter *error-symbols* nil)
(defmacro defunix-error (identifier)
  (check-type identifier keyword "a keyword identifier")
  (let ((cond-name (intern (symbol-name identifier)))
        (error-code
         (foreign-enum-value 'errno-values identifier)))
    `(progn
       (export ',cond-name)
       (define-condition ,cond-name (unix-error) ()
         (:default-initargs :code ,error-code
                            :identifier ,identifier))
       (push (cons ',identifier ',cond-name) *error-symbols*))))

(defmacro defunix-errors (symbols)
  `(progn
     ,@(loop :for symbol :in symbols
          :collect `(defunix-error ,symbol))))

(defunix-errors #.(loop :for k :being :the :hash-keys
                     :in (cffi::keyword-values
                          (cffi::parse-type 'et:errno-values))
                     :collect k))

(defun make-unix-error (err)
  (let* ((error-keyword
          (etypecase err
            (keyword err)
            (integer (or (ignore-errors
                           (foreign-enum-keyword 'errno-values err))
                         :unknown))))
         (error-code
          (etypecase err
            (keyword (ignore-errors
                       (foreign-enum-value 'errno-values error-keyword)))
            (integer err))))
    (let ((condition-class (cdr (assoc error-keyword *error-symbols*))))
      (if condition-class
          (make-condition condition-class)
          (make-condition 'unix-error
                          :code error-code
                          :identifier error-keyword)))))

(defmethod print-object ((unix-error unix-error) stream)
  (print-unreadable-object (unix-error stream :type t :identity nil)
    (let ((code (system-error-code unix-error))
          (identifier (system-error-identifier unix-error)))
      (format stream "~s ~s ~s"
              (or code "[No code]") identifier
              (or (ignore-errors (strerror (foreign-enum-keyword 'errno-values code)))
                  "[Can't get error string.]")))))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun foreign-name-to-symbol (s)
    (intern (substitute #\- #\_ (string-upcase s))))

  (defun choose-lisp-and-foreign-names (string-or-list)
    (etypecase string-or-list
      (string (values string-or-list (foreign-name-to-symbol string-or-list)))
      (list   (destructuring-bind (fname lname &rest args)
                  string-or-list
                (assert (stringp fname) (fname)
                        "Invalid name for alien function: ~A is not a string"
                        fname)
                (assert (symbolp lname) (lname)
                        "Invalid name for lisp function: ~A is not a symbol"
                        lname)
                (cond
                  ((member :wrap args)
                   ;; needed for function wrappers
                   (values (concatenate 'string fname "_wrap")
                           lname))
                  (t
                   (values fname lname))))))))

(defun signal-unix-error (&optional errcode)
  (declare (ignore errcode))
  (error (make-unix-error (get-errno))))

(defun never-fails (errcode)
  (declare (ignore errcode))
  nil)

(defmacro defwrapper (foreign-name lisp-name return-type docstring
                      error-predicate filter error-generator args)
  (labels ((canonical-type (type)
             (let ((norm-type (cffi::canonicalize-foreign-type type)))
               (if (and (keywordp type)
                        (member norm-type
                                '(:char :unsigned-char
                                  :short :unsigned-short
                                  :int :unsigned-int
                                  :long :unsigned-long
                                  :long-long :unsigned-long-long)))
                   norm-type
                   type)))
           (normalize-arg (arg)
             (list (first arg)
                   (canonical-type (second arg))))
           (invert-pair (pair)
             (list (second pair) (first pair))))

    (let* ((normalized-args (mapcar #'normalize-arg args))
           (arg-names (mapcar #'first normalized-args))
           (arglist (reduce #'append normalized-args :key #'invert-pair))
           (return-expression (if (eql filter 'identity) 'r `(,filter r))))
      (if (foreign-symbol-pointer foreign-name)
          `(defun ,lisp-name ,arg-names
             ,docstring
             ,(if (and (eql filter 'identity) (eql error-predicate 'never-fails))
                  `(foreign-funcall ,foreign-name
                                    ,@arglist
                                    ,(canonical-type return-type))
                  `(let ((r (foreign-funcall ,foreign-name
                                             ,@arglist
                                             ,(canonical-type return-type))))
                     ,(if (eql error-predicate 'never-fails)
                          `(progn ,return-expression)
                          `(if (,error-predicate r)
                               (,error-generator r)
                               ,return-expression)))))
          (error "Couldn't find definition for ~S" foreign-name)))))

(defmacro def-posix-call (foreign-name lisp-name return-type &rest argspecs)
  (let* ((first-argspec (car argspecs))
         (docstring (if (stringp first-argspec)
                        (progn
                          (setf argspecs (cdr argspecs))
                          first-argspec)
                        ""))
         (canonical-type (when (cffi::parse-type return-type)
                           (cffi::canonicalize
                            (cffi::parse-type return-type))))
         (error-predicate nil)
         (filter 'identity)
         (error-generator 'signal-unix-error))
    (case return-type
      ((:pointer)
       (setf error-predicate 'null-pointer-p))
      ((:string)
       (setf error-predicate '(lambda (s) (not (stringp s)))))
      ((:int :long ssize pid off)
       (setf error-predicate 'minusp))
      ;; FIXME: go here if the canonical type is unsigned
      ((:void :unsigned-int mode uid gid)
       (setf error-predicate 'never-fails))
      (t
       (cond
         ((eq canonical-type :pointer)
          (setf error-predicate 'null-pointer-p
                filter (iolib-alien-grovel:make-from-pointer-function-name
                        return-type)))
         (t (error "Could not choose a error-predicate function.")))))
    `(defwrapper ,foreign-name ,lisp-name ,return-type
                 ,docstring
                 ,error-predicate ,filter
                 ,error-generator
                 ,argspecs)))

(eval-when (:compile-toplevel :load-toplevel :execute)
  (defun process-cfun-argspecs (argspecs)
    (let ((docstring "")
          (export t))
      (when (eql (car argspecs) :noexport)
        (setf export nil)
        (setf argspecs (cdr argspecs)))
      ;; the docstring is optional
      (when (stringp (car argspecs))
        (setf docstring (car argspecs))
        (setf argspecs (cdr argspecs)))
      (values docstring export argspecs))))

(defmacro defsyscall (name return-type &rest args)
  (multiple-value-bind (docstring export argspecs) (process-cfun-argspecs args)
    (multiple-value-bind (foreign-name lisp-name)
        (choose-lisp-and-foreign-names name)
      (if export
          `(progn
             (export ',lisp-name)
             (declaim (inline ,lisp-name))
             (def-posix-call ,foreign-name ,lisp-name ,return-type
                             ,docstring ,@argspecs))
          `(progn
             (declaim (inline ,lisp-name))
             (def-posix-call ,foreign-name ,lisp-name ,return-type
                             ,docstring  ,@argspecs))))))

(defmacro defforeign (name return-type &rest args)
  (multiple-value-bind (docstring export argspecs) (process-cfun-argspecs args)
    (multiple-value-bind (foreign-name lisp-name)
        (choose-lisp-and-foreign-names name)
      (if export
          `(progn
             (export ',lisp-name)
             (declaim (inline ,lisp-name))
             (defwrapper ,foreign-name ,lisp-name ,return-type
                         ,docstring
                         never-fails identity identity
                         ,argspecs))
          `(progn
             (declaim (inline ,lisp-name))
             (defwrapper ,foreign-name ,lisp-name ,return-type
                         ,docstring
                         never-fails identity identity
                         ,argspecs))))))

(defmacro defforeign-with-helpers (name return-type &rest args)
  (let ((error-predicate-fun nil)
        (filter-fun nil)
        (error-generator-fun nil))
    (multiple-value-bind (docstring export argspecs) (process-cfun-argspecs args)
      (destructuring-bind (&key error-predicate filter error-generator)
          (car argspecs)
        (assert (and error-predicate filter error-generator))
        (setf error-predicate-fun error-predicate)
        (setf filter-fun filter)
        (setf error-generator-fun error-generator))
      (setf argspecs (cdr argspecs))
      (multiple-value-bind (foreign-name lisp-name)
          (choose-lisp-and-foreign-names name)
        (if export
            `(progn
               (export ',lisp-name)
               (declaim (inline ,lisp-name))
               (defwrapper ,foreign-name ,lisp-name ,return-type
                           ,docstring
                           ,error-predicate-fun ,filter-fun
                           ,error-generator-fun
                           ,argspecs))
            `(progn
               (declaim (inline ,lisp-name))
               (defwrapper ,foreign-name ,lisp-name ,return-type
                           ,docstring
                           ,error-predicate-fun ,filter-fun
                           ,error-generator-fun
                           ,argspecs)))))))

(defmacro define-entry-point (name args &body body)
  `(progn
     (export ',name)
     (declaim (inline ,name))
     (defun ,name ,args
       ,@body)))

;;
;; Errno & co.
;;

(defsyscall ("get_errno" get-errno :wrap) :int
  "Get current value of the errno variable.")

(defsyscall ("set_errno" set-errno :wrap) :int
  "Set current value of the errno variable."
  (value :int))

(defsyscall ("strerror_r" %strerror-r :wrap) :int :noexport
  "Thread-safe error string access."
  (errnum errno-values)
  (buf    :string)
  (size   size))

(define-entry-point strerror (errnum)
  "Return the string for UNIX error errnum."
  (with-foreign-pointer-as-string (buf 1024 bufsiz)
    (%strerror-r errnum buf bufsiz)))
